#include <net_funcs.h>
#include <utility.h>

volatile atomic_bool running = true; // NOLINT - not used unsafely, Drew

static void
signal_handler(int signal)
{
    if (signal == SIGINT)
    {
        if (false == running)
        {
            errno = ECANCELED;
            perror("\nHard Stopping");
            exit(0);
        }
        running = false;
    }
}

int
connect_to_server(char *dest, char *port)
{
    int              sock_fd      = 0;
    int              get_addr_res = 0;
    int              is_ipv4      = 0;
    struct addrinfo  hints        = { 0 };
    struct addrinfo *result       = NULL;
    struct addrinfo *cur_res      = NULL;
    // struct sockaddr_storage sockaddr     = { 0 }; // needed for inet_pton

    /*
    // set ai_family hint based on ip addr passed in
    // otherwise, it defaults to all configured ip's on the system
    // and may prefer a ipv6 over an ipv4, for example
    // pulled from man 3 inet_pton

    // Using gettaddr with AF_UNSPEC family hint is preferred
    if (1 == inet_pton(AF_INET6, dest, &sockaddr))
    {
        DEBUG_PRINT(("Valid IPv6 address; disabling name lookup\n"));
        hints.ai_family = AF_INET6;
        hints.ai_flags  = AI_NUMERICHOST;
    }
    else
    {
        DEBUG_PRINT(("Not a valid IPv6 address; assuming IPv4\n"));
        hints.ai_family = AF_INET;
    }
    */

    // set up hints to determine correct sockaddr style
    // pulled directly from man 3 getaddrinfo
    hints.ai_family   = AF_UNSPEC;   /* Accept IPv4/IPv6 */
    hints.ai_socktype = SOCK_STREAM; /* Datagram socket */
    hints.ai_protocol = 0;           /* Any protocol */

    // check results of getaddrinfo
    get_addr_res = getaddrinfo(dest, port, &hints, &result);
    if (0 != get_addr_res)
    {
        fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(get_addr_res));
        goto END;
    }

    /* getaddrinfo() returns a list of address structures.
                  Try each address until we successfully connect(2).
                  If socket(2) (or connect(2)) fails, we (close the socket
                  and) try the next address.
    Re-written from man page for BARR-C compliance (continue free)
    */
    for (cur_res = result; NULL != cur_res; cur_res = cur_res->ai_next)
    {
        sock_fd = socket(
            cur_res->ai_family, cur_res->ai_socktype, cur_res->ai_protocol);
        if (-1 != sock_fd)
        {
            if (-1 != connect(sock_fd, cur_res->ai_addr, cur_res->ai_addrlen))
            {
                DEBUG_PRINT(("Connected on fd: %d\n", sock_fd));
                break;
            }
        }
        close(sock_fd);
        sock_fd = -1; /* SUPER IMPORTANT, OTHERWISE SET FD ASSUMES IT WORKED */
    }

END:
    freeaddrinfo(result);
    result  = NULL;
    cur_res = NULL;
    return sock_fd;
}

size_t
send_file(int sockfd, FILE *fp, size_t fsize)
{
    size_t bytes_sent = 0;
    size_t to_send    = 0;
    char * tmp        = NULL;

    errno = 0; // reset errno

    // check value of tmp to ensure calloc worked as expected
    tmp = calloc(MAX_MSG_SIZE, sizeof(char));
    if (NULL == tmp)
    {
        goto CLEANUP;
    }

    // read up to MAX_MSG_SIZE from file, and store how many bytes were read
    // this also moves the FP forward
    // then, send the number of bytes read, and add to bytes_sent total
    while (bytes_sent < fsize)
    {
        to_send = fread(tmp, sizeof(char), MAX_MSG_SIZE, fp);
        bytes_sent += send(sockfd, tmp, to_send, 0);
        if (0 != errno)
        {
            perror("Error Sending File");
            goto CLEANUP;
        }
        DEBUG_PRINT(("Send %ld bytes\n", bytes_sent));
        memset(tmp, 0, MAX_MSG_SIZE); // NOLINT memset is fine, shut up - Drew
        if (0 == to_send)
        {
            break;
        }
    }

CLEANUP:
    free(tmp);
    tmp = NULL;
    return bytes_sent;
}

void *
interactive_recv(void *sockfd)
{
    bool          no_error = true;
    int           sfd      = *(int *)sockfd;
    size_t        num_recv = 0;
    size_t        to_recv  = 0;
    char          tmp      = '\0';
    struct pollfd fds[1]   = { 0 };

    /* watch socket for input */
    fds[0].fd     = sfd;
    fds[0].events = POLLIN;

    while (true == running)
    {
        poll(fds, 1, TIMEOUT);
        if (fds[0].revents)
        {
            num_recv = recv(sfd, &tmp, 1, 0);
            putc(tmp, stdout);
            fflush(stdout);
            if (0 == num_recv)
            {
                fprintf(stderr, "Broken Pipe; Stopping\n");
                running = false;
            }
        }
    }

    DEBUG_PRINT(("[+] running is false, stopping recv\n"));

END:
    return NULL;
}

void *
interactive_send(void *sockfd)
{
    bool          no_error = true;
    int           sfd      = *(int *)sockfd;
    size_t        to_send  = 0;
    char *        tmp      = NULL;
    struct pollfd fds[1]   = { 0 };

    /* watch stdin for input */
    fds[0].fd     = STDIN_FILENO;
    fds[0].events = POLLIN;

    // check value of tmp to ensure calloc worked as expected
    tmp = calloc(MAX_MSG_SIZE, sizeof(char));
    if (NULL == tmp)
    {
        goto END;
    }

    while (true == running)
    {
        poll(fds, 1, TIMEOUT);
        if (fds[0].revents)
        {
            to_send = read(STDIN_FILENO, tmp, MAX_MSG_SIZE);
            if (0 > send(sfd, tmp, to_send, 0))
            {
                fprintf(stderr, "Broken Pipe; Stopping\n");
                running = false;
            }
        }
    }

    DEBUG_PRINT(("[+] running is false, stoppping send\n"));

END:
    free(tmp);
    tmp = NULL;
    return NULL;
}

bool
interactive_wrapper(int sockfd)
{
    bool             no_error  = true;
    pthread_t        rtid      = 0;
    pthread_t        stid      = 0;
    struct sigaction sa_struct = { 0 };

    // Add SIGINT handler
    DEBUG_PRINT(("Adding SIGINT handler\n"));
    sa_struct.sa_handler = signal_handler;
    sigemptyset(&sa_struct.sa_mask);
    if (-1 == sigaction(SIGINT, &sa_struct, NULL))
    {
        perror("failed to create sigaction");
        goto ERROR;
    }

    if (0 != pthread_create(&rtid, NULL, interactive_recv, &sockfd))
    {
        perror("pthread_create() error");
        goto ERROR;
    }
    if (0 != pthread_create(&stid, NULL, interactive_send, &sockfd))
    {
        perror("pthread_create() error");
        goto ERROR;
    }

    if (0 != pthread_join(rtid, NULL))
    {
        fprintf(stderr, "Couldn't close read thread\n");
        goto ERROR;
    }
    if (0 != pthread_join(stid, NULL))
    {
        fprintf(stderr, "Couldn't close read thread\n");
        goto ERROR;
    }

    goto END;

ERROR:
    no_error = false;
END:
    return no_error;
}

bool
client(char *destination, char *port, char const *filename, bool inter_con)
{
    int         server_fd     = 0;
    size_t      bytes_sent    = 0;
    FILE *      infile_fp     = NULL;
    char *      file_contents = NULL;
    bool        did_error     = false;
    struct stat in_st         = { 0 };

    // reset errno
    errno = 0;

    // if filename was specified, get file stats and open for reading
    // pulled from man 3 fstatat
    if (NULL != filename)
    {
        if ((0 != stat(filename, &in_st)) || (0 != errno))
        {
            perror("In file error");
            did_error = true;
            goto CLEANUP;
        }

        infile_fp = fopen(filename, "re");
        if (NULL == infile_fp)
        {
            perror("Could not open file");
            did_error = true;
            goto CLEANUP;
        }

        DEBUG_PRINT(("%s has size %ld\n", filename, in_st.st_size));
    }

    // attempt to connect to server
    if (0 > (server_fd = connect_to_server(destination, port)))
    {
        did_error = true;
        fprintf(stderr, "Failed to Connect\n");
        goto CLEANUP;
    }

    // send file contents if file was specified and opend correctly
    // otherwise, begin threaded interactive session
    if (NULL != infile_fp)
    {
        DEBUG_PRINT(("Sending file %s\n", filename));
        bytes_sent = send_file(server_fd, infile_fp, in_st.st_size);
        if (in_st.st_size != bytes_sent)
        {
            fprintf(stderr,
                    "Failed to send full file; sent %ld bytes\n",
                    bytes_sent);
            did_error = true;
            goto CLEANUP;
        }

        if (true == inter_con)
        {
            if (false == interactive_wrapper(server_fd))
            {
                fprintf(stderr, "Error In Threaded Interactivity\n");
            }
        }
    }
    else
    {
        DEBUG_PRINT(("Entering Interactive Mode\n"));
        if (false == interactive_wrapper(server_fd))
        {
            fprintf(stderr, "Error In Threaded Interactivity\n");
        }
    }

CLEANUP:
    safe_close(server_fd);
    if (NULL != infile_fp)
    {
        fclose(infile_fp);
    }
    return did_error;
}
